<?php

declare(strict_types = 1);

/**
 * Copyright (C) 2024 PRONOVIX GROUP.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

namespace Drupal\symfony_webhook_receiver\DependencyInjection;

use Symfony\Component\DependencyInjection\ChildDefinition;
use Symfony\Component\DependencyInjection\Compiler\CompilerPassInterface;
use Symfony\Component\DependencyInjection\ContainerBuilder;
use Symfony\Component\DependencyInjection\Definition;
use Symfony\Component\RemoteEvent\Attribute\AsRemoteEventConsumer;
use Symfony\Component\RemoteEvent\Consumer\ConsumerInterface;

/**
 * Registers Remote Event Consumers via a compiler pass.
 *
 * This implementation was inspired by the Symfony Messenger + Drupal: Realtime
 * Queues and Cron (sm) module.
 *
 * @see https://drupal.org/project/sm.
 *
 * @internal This class is not part of the module's public API and is subject to change.
 */
final class RegisterAsRemoteEventConsumerAttributeCompilerPass implements CompilerPassInterface {

  /**
   * {@inheritdoc}
   */
  public function process(ContainerBuilder $container): void {
    // @phpstan-ignore argument.type
    foreach ($this->getRemoteEventConsumerClasses($container->getParameter('container.namespaces')) as $className) {
      // Don't create a service definition if this class is already a service.
      if ($container->hasDefinition($className)) {
        continue;
      }

      // registerAttributeForAutoconfiguration requires services to be created
      // beforehand. Symfony normally creates services for everything in src/.
      $definition = new Definition($className);
      $definition
        // Note: Autoconfigure does not work in services.yml file in <
        // Drupal 11.
        // @see https://www.drupal.org/project/drupal/issues/3221128
        ->setAutoconfigured(TRUE)
        ->setAutowired(TRUE)
        // Remote Event Consumers have to be public otherwise service locator
        // does not find them.
        ->setPublic(TRUE);

      $anonymousHash = ContainerBuilder::hash($className . mt_rand());
      $container->setDefinition('.' . $anonymousHash, $definition);
    }

    // Pulled from Symfony' FrameworkExtension.
    // https://github.com/symfony/framework-bundle/blob/v7.1.5/DependencyInjection/FrameworkExtension.php#L676-L678
    $container->registerAttributeForAutoconfiguration(AsRemoteEventConsumer::class, static function (ChildDefinition $definition, AsRemoteEventConsumer $attribute, \Reflector $reflector): void {
      $definition->addTag('remote_event.consumer', ['type' => $attribute->name]);

      // Drupal workaround starts here.
      assert($reflector instanceof \ReflectionClass);
      \preg_match_all('/^Drupal\\\\(?<moduleName>[a-z_0-9]{1,32})\\\\.*$/m', $reflector->name, $matches, PREG_SET_ORDER);
      $moduleName = $matches[0]['moduleName'] ?? throw new \Exception(\sprintf('Could not determine module name from class %s', $reflector->name));

      // \Drupal\Core\DependencyInjection\Compiler\LoggerAwarePass
      // autoconfigurator needs _provider.
      $definition->addTag('_provider', ['provider' => $moduleName]);
      // Drupal workaround ends here.
    });
  }

  /**
   * Get Remote Event Consumer classes for the provided namespaces.
   *
   * @param array<class-string, string> $namespaces
   *   An array of namespaces. Where keys are class strings and values are
   *   paths.
   *
   * @return \Generator<class-string>
   *   Generates class strings.
   *
   * @throws \ReflectionException
   */
  private function getRemoteEventConsumerClasses(array $namespaces): \Generator {
    foreach ($namespaces as $namespace => $dirs) {
      $dirs = (array) $dirs;
      foreach ($dirs as $dir) {
        $dir .= '/RemoteEvent';
        if (!file_exists($dir)) {
          continue;
        }
        $namespace .= '\\RemoteEvent';

        $iterator = new \RecursiveIteratorIterator(new \RecursiveDirectoryIterator($dir, \FilesystemIterator::KEY_AS_PATHNAME | \FilesystemIterator::CURRENT_AS_FILEINFO | \RecursiveDirectoryIterator::SKIP_DOTS), \RecursiveIteratorIterator::SELF_FIRST);
        foreach ($iterator as $fileinfo) {
          if ($fileinfo->getExtension() !== 'php') {
            continue;
          }

          /** @var \RecursiveDirectoryIterator|null $subDir */
          $subDir = $iterator->getSubIterator();
          if (NULL === $subDir) {
            continue;
          }

          $subDir = $subDir->getSubPath();
          $subDir = $subDir !== '' ? str_replace(DIRECTORY_SEPARATOR, '\\', $subDir) . '\\' : '';
          /** @var class-string $class */
          $class = $namespace . '\\' . $subDir . $fileinfo->getBasename('.php');

          $reflectionClass = new \ReflectionClass($class);

          if (!$reflectionClass->implementsInterface(ConsumerInterface::class)) {
            continue;
          }

          if (count($reflectionClass->getAttributes(AsRemoteEventConsumer::class)) > 0) {
            yield $class;
          }
        }
      }
    }
  }

}
